from django.contrib.auth.models import User
from django.test import TestCase
from django.utils.html import mark_safe

from squads.models import Groups, Memberships


class GroupsModelTest(TestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser", password="12345")

    def test_group_creation(self):
        group = Groups.objects.create(
            name="Test Group",
            owner=self.user,
            description="A test group",
        )
        self.assertEqual(group.name, "Test Group")
        self.assertEqual(group.owner, self.user)
        self.assertEqual(group.description, "A test group")
        self.assertTrue(group.is_active)
        self.assertFalse(group.require_approval)
        self.assertIsNotNone(group.created_at)
        self.assertIsNotNone(group.updated_at)

    def test_str_method(self):
        group = Groups.objects.create(name="Test Group", owner=self.user)
        self.assertEqual(str(group), "Test Group")

    def test_safe_description(self):
        group = Groups.objects.create(
            name="Test Group",
            owner=self.user,
            description='<script>alert("test")</script>',
        )
        self.assertEqual(group.safe_description(), mark_safe(group.description))


class MemberShipModelTest(TestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser", password="12345")

    def test_membership_creation(self):
        group = Groups.objects.create(name="Test Group", owner=self.user)
        membership = Memberships.objects.create(group=group, user=self.user)
        self.assertEqual(membership.group, group)
        self.assertEqual(membership.user, self.user)
        self.assertIsNotNone(membership.joined_at)

    def test_str_method(self):
        group = Groups.objects.create(name="Test Group", owner=self.user)
        membership = Memberships.objects.create(group=group, user=self.user)
        self.assertEqual(str(membership), f"{membership.user} in {membership.group}")



"""
Decorators
"""

import sys
from functools import wraps

from app_utils.esi import EsiDailyDowntime, fetch_esi_status

from squads.hooks import get_extension_logger

logger = get_extension_logger(__name__)

IS_TESTING = sys.argv[1:2] == ["test"]


def when_esi_is_available(func):
    """Make sure the decorated task only runs when esi is available.

    Raise exception when ESI is offline.
    Complete the task without running it when downtime is detected.

    Automatically disabled during tests.
    """

    @wraps(func)
    def outer(*args, **kwargs):
        if IS_TESTING is not True:
            try:
                fetch_esi_status().raise_for_status()
            except EsiDailyDowntime:
                logger.info("Daily Downtime detected. Aborting.")
                return None  # function will not run

        return func(*args, **kwargs)

    return outer



from typing import List

from ninja import NinjaAPI

from squads.api import schema
from squads.hooks import get_extension_logger
from squads.models.groups import Groups, Pending
from squads.models.memberships import Memberships

logger = get_extension_logger(__name__)


# pylint: disable=duplicate-code
class ManageApiEndpoints:
    tags = ["ManageGroups"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "groups/{group_id}/pendings/",
            response={200: List[schema.Pending], 403: str},
            tags=self.tags,
        )
        def get_pendings(request, group_id: int):
            groups = Groups.objects.visible_to(request.user)

            group_ids = [group.pk for group in groups]

            if group_id == 0:
                pending_groups = Pending.objects.filter(group__in=group_ids)
            else:
                pending_groups = Pending.objects.filter(group=group_id)

            output = []

            for pending in pending_groups:
                output.append(
                    {
                        "group": pending.group.name,
                        "group_id": pending.group.pk,
                        "user": pending.user.username,
                        "approved": pending.approved,
                        "created_at": pending.created_at,
                        "application": pending.application,
                        "application_id": pending.application_id,
                    }
                )

            return output

        @api.get(
            "groups/{group_id}/members/",
            response={200: List[schema.Members], 403: str},
            tags=self.tags,
        )
        def get_members(request, group_id: int):
            groups = Groups.objects.visible_to(request.user)

            group_ids = [group.pk for group in groups]

            if group_id == 0:
                membership_groups = Memberships.objects.filter(group__in=group_ids)
            else:
                membership_groups = Memberships.objects.filter(group=group_id)

            output = []

            for member in membership_groups:
                output.append(
                    {
                        "group": member.group.name,
                        "group_id": member.group.pk,
                        "user": member.user.username,
                        "has_required_skills": member.has_required_skills,
                        "is_active": member.is_active,
                        "joined_at": member.joined_at,
                        "application_id": member.application_id,
                    }
                )

            return output

        @api.get(
            "groups/{group_id}/",
            response={200: List[schema.Squads], 403: str},
            tags=self.tags,
        )
        def get_squads(request, group_id: int):
            groups = Groups.objects.visible_to(request.user)

            group_ids = [group.pk for group in groups]

            if group_id == 0:
                manage_groups = Groups.objects.filter(pk__in=group_ids)
            else:
                manage_groups = Groups.objects.filter(pk=group_id)

            output = []

            for group in manage_groups:
                output.append(
                    {
                        "group": group.name,
                        "group_id": group.pk,
                        "owner": group.owner.username,
                        "is_active": group.is_active,
                    }
                )

            return output



""" AA Hooks"""

import logging

from squads.app_settings import SQUADS_LOGGER_USE


def get_extension_logger(name):
    """
    Takes the name of a plugin/extension and generates a child logger of the extensions logger
    to be used by the extension to log events to the extensions logger.

    The logging level is determined by the level defined for the parent logger.

    :param: name: the name of the extension doing the logging
    :return: an extensions child logger
    """

    logger_name = "squads" if SQUADS_LOGGER_USE else "extensions"

    if not isinstance(name, str):
        raise TypeError(
            f"get_extension_logger takes an argument of type string."
            f"Instead received argument of type {type(name).__name__}."
        )

    parent_logger = logging.getLogger(logger_name)

    logger = logging.getLogger(logger_name + "." + name)
    logger.name = name
    logger.level = parent_logger.level

    return logger



"""App URLs"""

from django.urls import path, re_path
from django.views.static import serve

from squads.api import api
from squads.views.application import apply_group, cancel_group, leave_group
from squads.views.groups import broswe_groups, view_group
from squads.views.main import squads_index, squads_membership, squads_pending
from squads.views.manage import (
    accept_group,
    create_group,
    decline_group,
    delete_group,
    delete_membership,
    edit_group,
    manage_groups,
    manage_members,
    manage_pendings,
)

app_name: str = "squads"

urlpatterns = [
    # Main
    path("", squads_index, name="index"),
    path("browse", broswe_groups, name="groups"),
    path("create", create_group, name="create_group"),
    # Information
    path("pending", squads_pending, name="pending"),
    path("membership", squads_membership, name="membership"),
    # Manage
    path("accept/<str:application_id>", accept_group, name="accept_group"),
    path("decline/<str:application_id>", decline_group, name="decline_group"),
    path(
        "delete_membership/<str:application_id>",
        delete_membership,
        name="delete_membership",
    ),
    path("delete/<int:group_id>", delete_group, name="delete_group"),
    path("edit/<int:group_id>", edit_group, name="edit_group"),
    path("manage_pendings", manage_pendings, name="manage_pendings"),
    path("manage_members", manage_members, name="manage_members"),
    path("manage_groups", manage_groups, name="manage_groups"),
    # Application
    path("view/<int:group_id>", view_group, name="view_group"),
    path("apply/<int:group_id>", apply_group, name="apply_group"),
    path("cancel/<int:group_id>", cancel_group, name="cancel_group"),
    path("leave/<int:group_id>", leave_group, name="leave_group"),
    # API
    re_path(r"^api/", api.urls),
]
urlpatterns += [
    re_path(
        r"^groups_images/(?P<path>.*)$",
        serve,
        {
            "document_root": "squads/groups_images",
        },
    ),
]



from unittest.mock import MagicMock

from django.test import TestCase

from squads.auth_hooks import SquadsMenuItem
from squads.hooks import get_extension_logger


class TestTemplateTags(TestCase):
    def test_logger_fail(self):
        with self.assertRaises(TypeError):
            get_extension_logger(1234)


class AuthHooksTest(TestCase):
    def test_render_returns_empty_string_for_user_without_permission(self):
        # Setup
        squads_menu_item = SquadsMenuItem()
        mock_request = MagicMock()
        mock_request.user.has_perm.return_value = (
            False  # User does not have the required permission
        )

        # Action
        result = squads_menu_item.render(mock_request)

        # Assert
        self.assertEqual(
            result,
            "",
            "Expected render method to return an empty string for users without permission",
        )



"""
Manager for the Squad model.
"""

from django.db import models

from squads.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class GroupsQuerySet(models.QuerySet):
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all Squads for superuser %s.", user)
            return self

        if user.has_perm("squads.squad_admin"):
            logger.debug("Returning all Squads for Squad Admin %s.", user)
            return self

        if user.has_perm("squads.squad_manager"):
            query = models.Q(owner=user)
            logger.debug("Returning own Squads for %s.", user)
            return self.filter(query)
        return self.none()


class GroupsManager(models.Manager):
    def get_queryset(self):
        return GroupsQuerySet(self.model, using=self._db)

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)



"""Groups views."""

from django.contrib.auth.decorators import login_required, permission_required
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.shortcuts import get_object_or_404, redirect, render

from squads.forms import CommentForm
from squads.hooks import get_extension_logger
from squads.models.filters import SquadGroup
from squads.models.groups import Groups, Pending
from squads.models.memberships import Memberships
from squads.views._core import add_info_to_context

logger = get_extension_logger(__name__)


@login_required
@permission_required("squads.basic_access")
def broswe_groups(request):
    groups_list = Groups.objects.all()
    membership_ids = Memberships.objects.filter(user=request.user).values_list(
        "group_id", flat=True
    )
    paginator = Paginator(groups_list, 50)

    page = request.GET.get("page")
    try:
        groups = paginator.page(page)
    except PageNotAnInteger:
        groups = paginator.page(1)
    except EmptyPage:
        groups = paginator.page(paginator.num_pages)

    context = {
        "groups": groups,
        "membership_ids": membership_ids,
    }

    return render(
        request,
        "squads/groups/browse_groups.html",
        add_info_to_context(request, context),
    )


@login_required
@permission_required("squads.basic_access")
def view_group(request, group_id):
    group = get_object_or_404(Groups, id=group_id)
    comment_form = CommentForm(request.POST or None)
    pending_application = Pending.objects.filter(group=group, user=request.user).first()
    membership = Memberships.objects.filter(user=request.user, group=group).first()
    filters = SquadGroup.objects.filter(group=group).first()

    filter_req = True
    missing_req = []

    if filters:
        filter_req, missing_req = filters.check_user(request.user)

    if request.method == "POST" and "join_group" in request.POST:
        return redirect("squads:view_group", group_id=group_id)

    context = {
        "group": group,
        "comment_form": comment_form,
        "pending_application": pending_application,
        "membership": membership,
        "filter_req": filter_req,
        "missing_req": missing_req,
    }

    return render(
        request, "squads/groups/view_group.html", add_info_to_context(request, context)
    )
